"""
Create a new Project
"""
import os
import importlib
from .Base import Base
import sys

from pyPhases import Project, Data


class Run(Base):
    """create a Phase-Project"""

    config = None
    projectFileName = "project.yaml"
    packagePath = os.path.dirname(sys.modules["phases"].__file__)
    forceWrite = False
    debug = False
    phaseName = None

    def run(self):
        self.beforeRun()
        self.prepareConfig()

        project = self.createProjectFromConfig(self.config)

        self.runProject(project)

    def parseRunOptions(self):
        if self.options["-o"]:
            self.outputDir = self.options["-o"]
            sys.path.insert(0, self.outputDir)
            self.logDebug("Set Outputdir: %s" % (self.outputDir))
        if self.options["-p"]:
            self.projectFileName = self.options["-p"]
            self.logDebug("Set Projectfile: %s" % (self.projectFileName))
        if self.options["-c"]:
            self.projectConfigFileName = self.options["-c"]
            self.logDebug("Set Config file: %s" % (self.projectFileName))
        if self.options["-v"]:
            self.debug = True
        if "<phaseName>" in self.options:
            self.phaseName = self.options["<phaseName>"]

    def beforeRun(self):
        self.parseRunOptions()

    def getPackagePath(self, path):
        d = os.path.dirname(sys.modules["phases"].__file__)
        return os.path.join(d, path)

    def loadClass(self, classOptions, pythonPackage=""):
        name = classOptions["name"]
        path = classOptions["package"]
        options = classOptions["config"] if "config" in classOptions else {}
        leadingDot = "" if classOptions["system"] else "."
        package = None if classOptions["system"] else pythonPackage
        sys.path.insert(0, self.outputDir)
        module = importlib.import_module("%s%s.%s" % (leadingDot, path, name), package=package)
        classObj = getattr(module, name)

        if len(options) > 0:
            return classObj(options)
        else:
            return classObj()

    def getDataDefinition(self, dataObj):
        dependsOn = []
        if "dependsOn" in dataObj:
            for dependString in dataObj["dependsOn"]:
                dependsOn.append(dependString)

                if dependString not in self.config["config"] and dependString not in Data.dataNames:
                    self.logWarning(
                        "Dependency '%s' for Data could not be found in any config or other defined data" % (dependString)
                    )

        return Data(dataObj["name"], dependsOn)

    def createProjectFromConfig(self, config):
        self.logDebug("Load Project from Config")

        dataDefinitions = {}

        project = Project()
        project.debug = self.debug
        project.name = config["name"]
        project.namespace = config["namespace"]
        project.config = config["config"]

        for classObj in self.config["publisher"]:
            obj = self.loadClass(classObj, project.name)
            project.registerPublisher(obj)

        for classObj in self.config["exporter"]:
            obj = self.loadClass(classObj, project.name)
            project.registerExporter(obj)

        for classObj in self.config["storage"]:
            obj = self.loadClass(classObj, project.name)
            project.addStorage(obj)

        for stage in self.config["stages"]:
            project.addStage(stage)

        for dataObj in self.config["data"]:
            data = self.getDataDefinition(dataObj)
            dataDefinitions[dataObj["name"]] = data

        for phaseConfig in self.config["phases"]:
            obj = self.loadClass(phaseConfig, project.name)

            if not hasattr(obj, "exportData"):
                raise Exception(
                    "Phase %s was not initialized correctly, maybe you forgot to call the __init__ method after overwriting"
                )

            # add data
            if "exports" in phaseConfig:
                for dataName in phaseConfig["exports"]:
                    if dataName in dataDefinitions:
                        dataDef = dataDefinitions[dataName]
                    else:
                        dataDef = Data(dataName)

                    obj.exportData.append(dataDef)

            project.addPhase(obj, phaseConfig["stage"], phaseConfig["description"])

        return project

    def runProject(self, project):
        project.run(self.phaseName)



"""phases
Usage:
  phases [-v] ...
  phases create [-f] [-p <projectfile>] [-o <outputdir>] [-v]
  phases run [-p <projectfile>] [-c <configfile>] [-o <outputdir>] [-v]
  phases run <phaseName> [-p <projectfile>] [-c <configfile>] [-o <outputdir>] [-v]
  phases test [<testdir>] [-p <projectfile>] [-c <configfile>] [-o <outputdir>] [-tp <testpattern>] [-v]
  phases -h | --help
  phases --version
Options:
  -h --help                         Show this screen.
  --version                         Show version.
  -p <projectfile>                  project file [default: project.yaml]
  -o <outputdir>                    output directory [default: .]
  -f                                force to overwrite existing files
  -v                                verbose output
  -c <configfile>                   use userconfig (project.<configfile>.yaml)
  -t <testdir>                      test directory [default: tests]
  -tp <testpattern>                 pattern to look for tests [default: test.py]
Examples:
  phases --version
  phases create
  phases create -f
  phases create -f -o myOutput -p myProject.yaml
Help:
  For help using this tool, please open an issue on the Github repository:
"""

from inspect import getmembers, isclass
from docopt import docopt
from phases import __version__ as VERSION
import phases.commands
from phases.util.Logger import Logger, LogLevel
from phases.commands.Base import Base
import pyPhases

# import phases.commands as commands


def main():
    """Main CLI entrypoint."""
    options = docopt(__doc__, version=VERSION)

    if options["-v"]:
        Logger.verboseLevel = LogLevel.DEBUG
        pyPhases.util.Logger.verboseLevel = LogLevel.DEBUG

    Logger.log("Phases %s with pyPhases %s (Log level: %s)" % (VERSION, pyPhases.__version__, Logger.verboseLevel), "phases")

    # Here we'll try to dynamically match the command the user is trying to run
    # with a pre-defined command class we've already created.
    for (k, v) in options.items():
        if hasattr(phases.commands, k) and v:
            module = getattr(phases.commands, k)
            commands = getmembers(module, isclass)
            commands = list(filter(lambda c: c[0] != "Base" and issubclass(c[1], Base), commands))

            commandName, command = commands.pop()  # get last defined class to skip imports
            Logger.log("Run Command %s" % (commandName), "phases", LogLevel.DEBUG)
            command = command(options)
            command.run()



"""
Test a Project
"""

from unittest.result import TestResult
from pyPhases.test import TestCase, TestCaseIntegration
from unittest.loader import TestLoader
from unittest.suite import TestSuite
from unittest import TextTestRunner
from .run import Run
import os
import sys


class Test(Run):
    """test a Phase-Project by wrapping TestCases from pyPhase with the project and a configfile specified in project.test.yaml"""

    testDir = "tests"
    testPattern = "test*.py"

    def parseRunOptions(self):
        super().parseRunOptions()
        if self.options["<testdir>"]:
            self.testDir = os.path.join(self.outputDir, self.options["<testdir>"])
            sys.path.insert(0, self.testDir)
            self.logDebug("Set Testdir: %s" % (self.testDir))
        if self.options["<testpattern>"]:
            self.testPattern = self.options["<testpattern>"]
        if self.options["-c"] is None:
            self.projectConfigFileName = os.path.join(self.outputDir, "project.test.yaml")
            self.logDebug("Set Config file: %s" % (self.projectFileName))

    def wrapTestsInSuite(self, testOrSuite, wrapMethod):
        tests = []
        if issubclass(type(testOrSuite), TestSuite):
            for subTestOrSuite in testOrSuite._tests:
                tests += self.wrapTestsInSuite(subTestOrSuite, wrapMethod)
        else:
            check = wrapMethod(testOrSuite)
            if check:
                tests = [testOrSuite]
        return tests

    def run(self):
        self.beforeRun()
        self.prepareConfig()

        project = self.createProjectFromConfig(self.config)

        loader = TestLoader()
        self.logDebug("Discover Tests in %s for pattern %s (Basedir: %s)" % (self.testDir, self.testPattern, self.outputDir))
        os.chdir(self.outputDir)
        suite = loader.discover(self.testDir, self.testPattern)
        self.logDebug("Found Tests: %s" % (suite._tests))

        def wrapTestsInSuite(test):
            if isinstance(test, TestCaseIntegration):
                return False
            if isinstance(test, TestCase):
                test.setProject()
            return True

        TestCase.project = project
        noIntegrationTests = self.wrapTestsInSuite(suite, wrapTestsInSuite)

        suite = TestSuite()
        suite.addTests(noIntegrationTests)
        runner = TextTestRunner()
        return runner.run(suite)



class pdict(dict):
    def __getitem__(self, k):
        if not isinstance(k, list):
            return super().__getitem__(k)
        value = self
        for field in k:
            value = value.__getitem__(field)
        return value

    def __setitem__(self, k, v):
        if not isinstance(k, list):
            return super().__setitem__(k, v)
        value = self
        overwriteField = k.pop()
        for field in k:
            value = value.__getitem__(field)
        value[overwriteField] = v



import setuptools

from phases import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="phases",
    version=__version__,
    author="Franz Ehrlich",
    author_email="fehrlichd@gmail.com",
    description="A Framework for creating a boilerplate template for ai projects that are ready for MLOps",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/tud.ibmt/phases",
    packages=setuptools.find_packages(exclude=["tests", "example"]),
    install_requires = ['docopt', 'pyPhases', 'PyYAML', 'pystache'],
    include_package_data=True,
    package_data={'phases': ['generate-template/*','generate-template/**/*', 'static-template/*','static-template/**/*']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points = {
        'console_scripts': [
            'phases=phases.cli:main',
        ],
    },
)



from phases.util.pdict import pdict
import yaml
import os
import pyPhases
import sys
from benedict import benedict
from ..util.Logger import classLogger


class Misconfigured(Exception):
    pass


@classLogger
class Base(object):
    outputDir = None
    """A base command."""

    def __init__(self, options, *args, **kwargs):
        self.options = options
        self.args = args
        self.kwargs = kwargs
        self.config = {}
        self.projectFileName = "project.yaml"
        self.projectConfigFileName = None
        self.config = None

    def run(self):
        raise NotImplementedError("You must implement the run() method yourself!")

    def overwriteConfig(self, config, values, path=[]):
        config = pdict(config)
        for field in values:
            value = values[field]
            p = path + [field]

            if isinstance(value, dict):
                config[p] = self.overwriteConfig(config[p], value, [])
            else:
                config[p] = values[field]

        return config

    def prepareConfig(self):
        projectFile = open(self.projectFileName, "r")
        yamlContent = projectFile.read()
        projectConfig = yaml.load(yamlContent, Loader=yaml.SafeLoader)

        if self.projectConfigFileName is not None:
            self.logDebug("set user config: %s" % self.projectConfigFileName)
            yamlContent = open(self.projectConfigFileName, "r").read()
            userConfig = yaml.load(yamlContent, Loader=yaml.SafeLoader)
            startPath = [] if "isFullConfig" in userConfig and userConfig["isFullConfig"] else ["config"]
            projectConfig = self.overwriteConfig(projectConfig, userConfig, startPath)

        self.validateConfig(projectConfig)
        if self.outputDir == None:
            self.outputDir = projectConfig["name"]
        self.config = projectConfig
        self.normalizeConfigArrays()

        self.preparePhases()
        self.preparePackageConfig()

    def validateConfig(self, config):
        # check required fields
        required = ["name", "namespace", "phases"]
        for field in required:
            if not field in config:
                raise Misconfigured(u"%s is required in the project yaml file" % (field))

        # set default values
        defaultValues = {
            "storage": ["FileStorage"],
            "publisher": [],
            "exporter": ["ObjectExporter"],
            "usedClasses": [],
        }
        # for field in defaultValues:
        for field in defaultValues:
            if not field in config:
                config[field] = defaultValues[field]

    def registerClassName(self, classObj, package):
        className = classObj["name"]
        ucFirstClassname = className[0].upper() + className[1:]
        userClassName = "user" + ucFirstClassname
        systemClassName = "system" + ucFirstClassname

        self.config[userClassName] = []
        self.config[systemClassName] = []

        systemclass = self.checkIfClassExistsInPyPhases(className, package)
        if systemclass:
            packagePre = "pyPhases."
        else:
            packagePre = ""

        classObj["package"] = packagePre + package
        classObj["packageName"] = package
        classObj["system"] = systemclass

        if systemclass:
            self.config[systemClassName].append(classObj)
        else:
            self.config[userClassName].append(classObj)
        self.config["usedClasses"].append(classObj)
        return classObj

    def preparePackageConfig(self):
        self.config["packages"] = []
        packages = {}
        for classObj in self.config["usedClasses"]:
            if not classObj["system"]:
                if classObj["packageName"] in packages:
                    packages[classObj["packageName"]]["imports"].append(classObj["name"])
                else:
                    packages[classObj["packageName"]] = {"name": classObj["packageName"], "imports": [classObj["name"]]}
        for name in packages:
            self.config["packages"].append(packages[name])

    def preparePhases(self):
        # flatten phases and extract stages
        self.config["stages"] = []
        self.config["userPhases"] = []
        self.config["phaseClasses"] = []
        lastPhase = None
        for stageOrPhase in self.config["phases"]:
            isPhase = type(stageOrPhase) == dict
            if isPhase:
                stageName = "default"
                phases = [stageOrPhase]
            else:
                stageName = stageOrPhase
                phases = self.config["stages"]["phases"]

            if stageName not in self.config["stages"]:
                self.config["stages"].append(stageName)

            for phase in phases:
                phase["stage"] = stageName
                self.config["userPhases"].append(phase)
                self.registerClassName(phase, "phases")
                if lastPhase != None:
                    lastPhase["next"] = phase
                lastPhase = phase

        self.config["phases"] = self.config["userPhases"]

    def normalizeDataArray(self, objectOrString):
        if isinstance(objectOrString, str):
            return {"name": objectOrString, "description": ""}

        if not "name" in objectOrString:
            raise Exception("One Object does not have a name")

        if "config" in objectOrString:
            objectOrString["config"]["items"] = self.normalizeDict(objectOrString["config"])

        return objectOrString

    def normalizeDict(self, dictObj):
        arrayForTemplate = []

        for (__, name) in enumerate(dictObj):
            arrayForTemplate.append({"name": name, "value": dictObj[name]})
        return arrayForTemplate

    def normalizeConfigArrays(self):
        arrayFields = ["publisher", "storage", "exporter", "data"]
        for field in arrayFields:
            if field not in self.config:
                self.config[field] = []
            for index, arrayOrString in enumerate(self.config[field]):
                obj = self.normalizeDataArray(arrayOrString)
                obj = self.registerClassName(obj, field)
                self.config[field][index] = obj

    def getPackagePath(self, path):
        d = os.path.dirname(sys.modules["phases"].__file__)
        return os.path.join(d, path)

    def checkIfClassExistsInPyPhases(self, className, package):
        try:
            module = getattr(pyPhases, package)
            getattr(module, className)
            return True
        except AttributeError:
            pass

        return False



"""
Create a new Project
"""

import os
from pathlib import Path
from .Base import Base
import sys
from distutils.dir_util import copy_tree

import pystache

class Create(Base):
    """create a Phase-Project"""

    templateDir = "generate-template/"
    staticFilesDir = "static-template/"
    packagePath = os.path.dirname(sys.modules["phases"].__file__)
    forceWrite = False

    def parseRunOptions(self):
        if (self.options['-o']):
            self.outputDir = self.options['-o']
            self.logDebug("Set Outputdir: %s" % (self.outputDir))
        if (self.options['-p']):
            self.projectFileName = self.options['-p']
            self.logDebug("Set Projectfile: %s" % (self.projectFileName))
        if (self.options['-f']):
            self.forceWrite = True

    def beforeRun(self):
        self.parseRunOptions()
        self.templateDir = self.getPackagePath(self.templateDir)
        self.staticFilesDir = self.getPackagePath(self.staticFilesDir)

    def run(self):
        self.beforeRun()
        self.prepareConfig()

        self.copyStaticFiles()
        self.generateInitFiles()
        self.generateFilesFromConfig()

    def copyStaticFiles(self):
        self.log("Copy static files")
        self.logDebug("Copy static files from %s to %s" %(self.staticFilesDir, self.outputDir))

        copy_tree(self.staticFilesDir, self.outputDir)

    def generateInitFiles(self):
        # this method assumes the python files will be saved in {{name}}
        self.log("Generate init Files")
        for package in self.config["packages"]:
            templateFile = open(self.templateDir + "packages.m", "r")
            templateContent = templateFile.read()

            path = self.outputDir + '/' + self.config["name"] + "/" + package["name"] + "/"

            self.writeTemplateFile(path, '__init__.py', templateContent,
                                   package)

    def generateFilesFromConfig(self):
        self.log("Generate files for the project with template files")
        outputDir = self.outputDir
        # iterate all template files (*.m files are for manual use)
        pathlist = Path(self.templateDir).glob('**/*.mustache')
        for path in pathlist:
            templateFile = open(path, "r")
            templateContent = templateFile.read()
            pathDiff = os.path.relpath(path.parent, Path(self.templateDir))

            fileName = pystache.render(path.stem, self.config)
            pathDiff = pystache.render(pathDiff, self.config)
            relPath = os.path.join(outputDir, pathDiff)

            self.logDebug("write template %s to path to %s" %
                          (path.stem, relPath))

            if (not fileName in self.config):
                self.writeTemplateFile(relPath, fileName, templateContent)
            else:
                for obj in self.config[fileName]:
                    if (not obj["system"]):
                        self.log("Create stub for %s/%s" %
                                 (fileName, obj["name"]))
                        self.writeTemplateFile(relPath + '/' + fileName,
                                               obj["name"] + '.py',
                                               templateContent, obj)

    def writeTemplateFile(self, path, filename, tplString,
                          templateObject=None):
        if templateObject == None:
            templateObject = self.config

        if not os.path.exists(path):
            os.makedirs(path)

        fullPath = path + '/' + filename
        self.logDebug("Write file %s" % (fullPath))
        if self.forceWrite or not os.path.isfile(fullPath):
            fileContent = pystache.render(tplString, templateObject)
            file = open(fullPath, "w")
            file.write(fileContent)
            file.close()
        else:
            self.logWarning(
                "File %s allready exists and was not overwritten, use the option -f to force to overwrite it "
                % (fullPath))



from enum import Enum
from .Singleton import Singleton
from functools import partial, wraps


class LogLevel(Enum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3


class Logger():
    verboseLevel: LogLevel = LogLevel.INFO

    def log(msg, system=None, level=LogLevel.INFO):
        if (system != None):
            msg = "[" + system + "] " + msg

        if (level == LogLevel.WARNING):
            msg = u"\033[33;1;4m%s\033[0m" % (msg)
        if (level == LogLevel.ERROR):
            msg = u"\033[31;1;4m%s\033[0m" % (msg)

        if (Logger.verboseLevel.value <= level.value):
            print(msg)


def classLogger(class_):
    name = class_.__name__

    def log(self, msg, level=LogLevel.INFO):
        Logger.log(msg, name, level)

    def logDebug(self, msg):
        Logger.log(msg, name, level=LogLevel.DEBUG)

    def logWarning(self, msg):
        Logger.log(msg, name, LogLevel.WARNING)

    class_.log = log
    class_.logDebug = logDebug
    class_.logWarning = logWarning
    return class_